import pytest


@pytest.mark.skip
def test_relu_clips_negative(model, input_BI):
    # TODO: There was a bug related to this
    pass


@pytest.mark.skip
def test_relu_clipped_gradients_are_zero():
    # TODO: There was a bug related to this
    pass
